@@ -16,8 +16,8 @@
  */
 package org.apache.qpid.jms;
 
-import java.io.IOException;
 import java.net.URI;
+import java.net.URISyntaxException;
 import java.util.Map;
 import java.util.Set;
 import java.util.concurrent.ConcurrentHashMap;
@@ -38,6 +38,7 @@ import javax.jms.IllegalStateException;
 import javax.jms.InvalidClientIDException;
 import javax.jms.InvalidDestinationException;
 import javax.jms.JMSException;
+import javax.jms.JMSRuntimeException;
 import javax.jms.Queue;
 import javax.jms.QueueConnection;
 import javax.jms.QueueSession;
@@ -54,6 +55,7 @@ import org.apache.qpid.jms.exceptions.JmsExceptionSupport;
 import org.apache.qpid.jms.message.JmsInboundMessageDispatch;
 import org.apache.qpid.jms.message.JmsMessage;
 import org.apache.qpid.jms.message.JmsMessageFactory;
+import org.apache.qpid.jms.message.JmsMessageTransformation;
 import org.apache.qpid.jms.message.JmsOutboundMessageDispatch;
 import org.apache.qpid.jms.meta.JmsConnectionId;
 import org.apache.qpid.jms.meta.JmsConnectionInfo;
@@ -74,13 +76,18 @@ import org.apache.qpid.jms.policy.JmsPresettlePolicy;
 import org.apache.qpid.jms.policy.JmsRedeliveryPolicy;
 import org.apache.qpid.jms.provider.AsyncResult;
 import org.apache.qpid.jms.provider.Provider;
-import org.apache.qpid.jms.provider.ProviderClosedException;
 import org.apache.qpid.jms.provider.ProviderConstants.ACK_TYPE;
+import org.apache.qpid.jms.provider.ProviderException;
 import org.apache.qpid.jms.provider.ProviderFuture;
 import org.apache.qpid.jms.provider.ProviderListener;
 import org.apache.qpid.jms.provider.ProviderSynchronization;
+import org.apache.qpid.jms.tracing.JmsTracer;
+import org.apache.qpid.jms.util.FifoMessageQueue;
+import org.apache.qpid.jms.util.MessageQueue;
+import org.apache.qpid.jms.util.PriorityMessageQueue;
 import org.apache.qpid.jms.util.QpidJMSThreadFactory;
 import org.apache.qpid.jms.util.ThreadPoolUtils;
+import org.apache.qpid.jms.util.URISupport;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
@@ -91,12 +98,13 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
 
     private static final Logger LOG = LoggerFactory.getLogger(JmsConnection.class);
 
-    private final Map<JmsSessionId, JmsSession> sessions = new ConcurrentHashMap<JmsSessionId, JmsSession>();
+    private final Map<JmsSessionId, JmsSession> sessions = new ConcurrentHashMap<>();
+    private final Map<JmsConsumerId, JmsConnectionConsumer> connectionConsumers = new ConcurrentHashMap<>();
     private final AtomicBoolean connected = new AtomicBoolean();
     private final AtomicBoolean closed = new AtomicBoolean();
     private final AtomicBoolean closing = new AtomicBoolean();
     private final AtomicBoolean started = new AtomicBoolean();
-    private final AtomicReference<IOException> failureCause = new AtomicReference<>();
+    private final AtomicReference<Exception> failureCause = new AtomicReference<>();
     private final JmsConnectionInfo connectionInfo;
     private final ThreadPoolExecutor executor;
 
@@ -104,15 +112,13 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
     private JmsMessageFactory messageFactory;
     private Provider provider;
 
-    private final Set<JmsConnectionListener> connectionListeners =
-        new CopyOnWriteArraySet<JmsConnectionListener>();
-    private final Map<JmsTemporaryDestination, JmsTemporaryDestination> tempDestinations =
-        new ConcurrentHashMap<JmsTemporaryDestination, JmsTemporaryDestination>();
+    private final Set<JmsConnectionListener> connectionListeners = new CopyOnWriteArraySet<>();
+    private final Map<JmsTemporaryDestination, JmsTemporaryDestination> tempDestinations = new ConcurrentHashMap<>();
     private final AtomicLong sessionIdGenerator = new AtomicLong();
     private final AtomicLong tempDestIdGenerator = new AtomicLong();
     private final AtomicLong transactionIdGenerator = new AtomicLong();
-
-    private final Map<AsyncResult, AsyncResult> requests = new ConcurrentHashMap<AsyncResult, AsyncResult>();
+    private final AtomicLong connectionConsumerIdGenerator = new AtomicLong();
+    private final Map<AsyncResult, AsyncResult> requests = new ConcurrentHashMap<>();
 
     protected JmsConnection(final JmsConnectionInfo connectionInfo, Provider provider) throws JMSException {
 
@@ -144,6 +150,7 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         }
 
         this.connectionInfo = connectionInfo;
+        this.connectionInfo.setConnection(this);
     }
 
     JmsConnection connect() throws JMSException {
@@ -154,7 +161,9 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         try {
             provider.connect(connectionInfo);
         } catch (Exception ex) {
-            LOG.error("Failed to connect to remote at: {}", connectionInfo.getConfiguredURI());
+            URI remoteURI = connectionInfo.getConfiguredURI();
+            LOG.error("Failed to connect to remote at: {}",
+                remoteURI.getScheme() + "://" + remoteURI.getHost() + ":" + remoteURI.getPort());
             LOG.trace("Error: ", ex);
             try {
                 provider.close();
@@ -202,8 +211,12 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
                     session.shutdown();
                 }
 
+                for (JmsConnectionConsumer connectionConsumer : connectionConsumers.values()) {
+                    connectionConsumer.shutdown();
+                }
+
                 if (isConnected() && !isFailed()) {
-                    ProviderFuture request = new ProviderFuture();
+                    ProviderFuture request = provider.newProviderFuture();
                     requests.put(request, request);
                     try {
                         provider.destroy(connectionInfo, request);
@@ -218,8 +231,8 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
                             }
                             LOG.debug("Failed destroying Connection resource: {}", ex.getMessage());
                         }
-                    } catch(ProviderClosedException pce) {
-                        LOG.debug("Ignoring provider closed exception during connection close");
+                    } catch(ProviderException prodiverError) {
+                        LOG.debug("Ignoring provider exception during connection close");
                     } finally {
                         requests.remove(request);
                     }
@@ -231,6 +244,7 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
                 started.set(false);
                 closing.set(false);
                 closed.set(true);
+                connectionInfo.getTracer().close();
             }
         } catch (Exception e) {
             throw JmsExceptionSupport.create(e);
@@ -271,6 +285,10 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
             session.shutdown(cause);
         }
 
+        for (JmsConnectionConsumer connectionConsumer : connectionConsumers.values()) {
+            connectionConsumer.shutdown();
+        }
+
         if (isConnected() && !isFailed() && !closing.get()) {
             destroyResource(connectionInfo);
         }
@@ -296,7 +314,6 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         createJmsConnection();
         int ackMode = getSessionAcknowledgeMode(transacted, acknowledgeMode);
         JmsSession result = new JmsSession(this, getNextSessionId(), ackMode);
-        addSession(result.getSessionInfo(), result);
         if (started.get()) {
             result.start();
         }
@@ -342,9 +359,13 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         createJmsConnection();
         if (started.compareAndSet(false, true)) {
             try {
-                for (JmsSession s : sessions.values()) {
-                    s.start();
+                for (JmsSession session : sessions.values()) {
+                    session.start();
                 }
+
+                for (JmsConnectionConsumer connectionConsumer : connectionConsumers.values()) {
+                    connectionConsumer.start();
+                 }
             } catch (Exception e) {
                 throw JmsExceptionSupport.create(e);
             }
@@ -376,9 +397,15 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         }
 
         if (started.compareAndSet(true, false)) {
-            synchronized(sessions) {
-                for (JmsSession s : sessions.values()) {
-                    s.stop();
+            synchronized (sessions) {
+                for (JmsSession session : sessions.values()) {
+                    session.stop();
+                }
+            }
+
+            synchronized (connectionConsumers) {
+                for (JmsConnectionConsumer connectionConsumer : connectionConsumers.values()) {
+                    connectionConsumer.stop();
                 }
             }
         }
@@ -388,48 +415,95 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
     public ConnectionConsumer createSharedConnectionConsumer(Topic topic, String subscriptionName, String messageSelector, ServerSessionPool sessionPool, int maxMessages) throws JMSException {
         checkClosedOrFailed();
         createJmsConnection();
-        throw new JMSException("Not supported");
+
+        return createConnectionConsumer(topic, messageSelector, sessionPool, maxMessages, subscriptionName, false, true);
     }
 
     @Override
     public ConnectionConsumer createSharedDurableConnectionConsumer(Topic topic, String subscriptionName, String messageSelector, ServerSessionPool sessionPool, int maxMessages) throws JMSException {
         checkClosedOrFailed();
         createJmsConnection();
-        throw new JMSException("Not supported");
+
+        return createConnectionConsumer(topic, messageSelector, sessionPool, maxMessages, subscriptionName, true, true);
     }
 
     @Override
-    public ConnectionConsumer createDurableConnectionConsumer(Topic topic, String subscriptionName,
-                                                              String messageSelector, ServerSessionPool sessionPool, int maxMessages) throws JMSException {
+    public ConnectionConsumer createDurableConnectionConsumer(Topic topic, String subscriptionName, String messageSelector, ServerSessionPool sessionPool, int maxMessages) throws JMSException {
         checkClosedOrFailed();
         createJmsConnection();
-        throw new JMSException("Not supported");
+
+        return createConnectionConsumer(topic, messageSelector, sessionPool, maxMessages, subscriptionName, true, false);
     }
 
     @Override
-    public ConnectionConsumer createConnectionConsumer(Destination destination, String messageSelector,
-                                                       ServerSessionPool sessionPool, int maxMessages) throws JMSException {
+    public ConnectionConsumer createConnectionConsumer(Destination destination, String messageSelector, ServerSessionPool sessionPool, int maxMessages) throws JMSException {
         checkClosedOrFailed();
         createJmsConnection();
-        throw new JMSException("Not supported");
+
+        return createConnectionConsumer(destination, messageSelector, sessionPool, maxMessages, null, false, false);
     }
 
     @Override
-    public ConnectionConsumer createConnectionConsumer(Topic topic, String messageSelector,
-                                                       ServerSessionPool sessionPool, int maxMessages) throws JMSException {
+    public ConnectionConsumer createConnectionConsumer(Topic topic, String messageSelector, ServerSessionPool sessionPool, int maxMessages) throws JMSException {
         checkClosedOrFailed();
         createJmsConnection();
-        throw new JMSException("Not supported");
+
+        return createConnectionConsumer(topic, messageSelector, sessionPool, maxMessages, null, false, false);
     }
 
     @Override
-    public ConnectionConsumer createConnectionConsumer(Queue queue, String messageSelector,
-                                                       ServerSessionPool sessionPool, int maxMessages) throws JMSException {
+    public ConnectionConsumer createConnectionConsumer(Queue queue, String messageSelector, ServerSessionPool sessionPool, int maxMessages) throws JMSException {
         checkClosedOrFailed();
         createJmsConnection();
-        throw new JMSException("Not supported");
+
+        return createConnectionConsumer(queue, messageSelector, sessionPool, maxMessages, null, false, false);
     }
 
+    private ConnectionConsumer createConnectionConsumer(Destination destination, String messageSelector, ServerSessionPool sessionPool, int maxMessages, String subscriptionName, boolean durable, boolean shared) throws JMSException {
+        JmsDestination jmsDestination = JmsMessageTransformation.transformDestination(this, destination);
+
+        int configuredPrefetch = getPrefetchPolicy().getConfiguredPrefetch((JmsSession) null, jmsDestination, durable, false);
+
+        final MessageQueue messageQueue;
+
+        if (isLocalMessagePriority()) {
+            messageQueue = new PriorityMessageQueue();
+        } else {
+            messageQueue = new FifoMessageQueue(configuredPrefetch);
+        }
+
+        JmsConsumerInfo consumerInfo = new JmsConsumerInfo(getNextConnectionConsumerId(), null);
+        consumerInfo.setExplicitClientID(isExplicitClientID());
+        consumerInfo.setSelector(messageSelector);
+        consumerInfo.setDurable(durable);
+        consumerInfo.setSubscriptionName(subscriptionName);
+        consumerInfo.setShared(shared);
+        consumerInfo.setDestination(jmsDestination);
+        consumerInfo.setAcknowledgementMode(Session.AUTO_ACKNOWLEDGE);
+        consumerInfo.setNoLocal(false);
+        consumerInfo.setBrowser(false);
+        consumerInfo.setPrefetchSize(configuredPrefetch);
+        consumerInfo.setRedeliveryPolicy(getRedeliveryPolicy().copy());
+        consumerInfo.setLocalMessageExpiry(isLocalMessageExpiry());
+        consumerInfo.setPresettle(false);
+        consumerInfo.setDeserializationPolicy(getDeserializationPolicy().copy());
+        consumerInfo.setMaxMessages(maxMessages);
+        consumerInfo.setConnectionConsumer(true);
+
+        JmsConnectionConsumer consumer = new JmsConnectionConsumer(this, consumerInfo, messageQueue, sessionPool);
+
+        try {
+            consumer.init();
+            if (started.get()) {
+                consumer.start();
+            }
+            return consumer;
+        } catch (JMSException jmsEx) {
+            consumer.close();
+            throw jmsEx;
+        }
+     }
+
     @Override
     public TopicSession createTopicSession(boolean transacted, int acknowledgeMode) throws JMSException {
         checkClosedOrFailed();
@@ -475,14 +549,18 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
 
         if (transacted) {
             result = Session.SESSION_TRANSACTED;
-        } else if (acknowledgeMode < Session.SESSION_TRANSACTED || acknowledgeMode > Session.DUPS_OK_ACKNOWLEDGE){
-            throw new JMSException("acknowledgeMode " + acknowledgeMode + " cannot be used for an non-transacted Session");
+        } else {
+            try {
+                JmsSession.validateSessionMode(acknowledgeMode);
+            } catch (JMSRuntimeException jmsre) {
+                throw new JMSException("acknowledgeMode " + acknowledgeMode + " cannot be used for an non-transacted Session");
+            }
         }
 
         return result;
     }
 
-    protected void removeSession(JmsSessionInfo sessionInfo) throws JMSException {
+    protected void removeSession(JmsSessionInfo sessionInfo) {
         sessions.remove(sessionInfo.getId());
     }
 
@@ -490,6 +568,14 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         sessions.put(sessionInfo.getId(), session);
     }
 
+    protected void removeConnectionConsumer(JmsConsumerInfo consumerInfo) {
+        connectionConsumers.remove(consumerInfo.getId());
+    }
+
+    protected void addConnectionConsumer(JmsConsumerInfo consumerInfo, JmsConnectionConsumer consumer) {
+        connectionConsumers.put(consumerInfo.getId(), consumer);
+    }
+
     private void createJmsConnection() throws JMSException {
         if (isConnected() || closed.get()) {
             return;
@@ -582,6 +668,10 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         return new JmsTransactionId(connectionInfo.getId(), transactionIdGenerator.incrementAndGet());
     }
 
+    protected JmsConsumerId getNextConnectionConsumerId() {
+        return new JmsConsumerId(connectionInfo.getId().toString(), -1, connectionConsumerIdGenerator.incrementAndGet());
+    }
+
     protected synchronized boolean isExplicitClientID() {
         return connectionInfo.isExplicitClientID();
     }
@@ -596,7 +686,7 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         checkClosedOrFailed();
 
         try {
-            ProviderFuture request = new ProviderFuture(synchronization);
+            ProviderFuture request = provider.newProviderFuture(synchronization);
             requests.put(request, request);
             try {
                 provider.create(resource, request);
@@ -617,7 +707,7 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         checkClosedOrFailed();
 
         try {
-            ProviderFuture request = new ProviderFuture(synchronization);
+            ProviderFuture request = provider.newProviderFuture(synchronization);
             requests.put(request, request);
             try {
                 provider.start(resource, request);
@@ -638,7 +728,7 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         checkClosedOrFailed();
 
         try {
-            ProviderFuture request = new ProviderFuture(synchronization);
+            ProviderFuture request = provider.newProviderFuture(synchronization);
             requests.put(request, request);
             try {
                 provider.stop(resource, request);
@@ -659,7 +749,7 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         checkClosedOrFailed();
 
         try {
-            ProviderFuture request = new ProviderFuture(synchronization);
+            ProviderFuture request = provider.newProviderFuture(synchronization);
             requests.put(request, request);
             try {
                 provider.destroy(resource, request);
@@ -672,15 +762,11 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         }
     }
 
-    void send(JmsOutboundMessageDispatch envelope) throws JMSException {
-        send(envelope, null);
-    }
-
     void send(JmsOutboundMessageDispatch envelope, ProviderSynchronization synchronization) throws JMSException {
         checkClosedOrFailed();
 
         try {
-            ProviderFuture request = new ProviderFuture(synchronization);
+            ProviderFuture request = provider.newProviderFuture(synchronization);
             requests.put(request, request);
             try {
                 provider.send(envelope, request);
@@ -701,7 +787,7 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         checkClosedOrFailed();
 
         try {
-            ProviderFuture request = new ProviderFuture(synchronization);
+            ProviderFuture request = provider.newProviderFuture(synchronization);
             provider.acknowledge(envelope, ackType, request);
             request.sync();
         } catch (Exception ioe) {
@@ -717,7 +803,7 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         checkClosedOrFailed();
 
         try {
-            ProviderFuture request = new ProviderFuture(synchronization);
+            ProviderFuture request = provider.newProviderFuture(synchronization);
             provider.acknowledge(sessionId, ackType, request);
             request.sync();
         } catch (Exception ioe) {
@@ -733,7 +819,7 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         checkClosedOrFailed();
 
         try {
-            ProviderFuture request = new ProviderFuture(synchronization);
+            ProviderFuture request = provider.newProviderFuture(synchronization);
             requests.put(request, request);
             try {
                 provider.unsubscribe(name, request);
@@ -754,7 +840,7 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         checkClosedOrFailed();
 
         try {
-            ProviderFuture request = new ProviderFuture(synchronization);
+            ProviderFuture request = provider.newProviderFuture(synchronization);
             requests.put(request, request);
             try {
                 provider.commit(transactionInfo, nextTransactionId, request);
@@ -775,7 +861,7 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         checkClosedOrFailed();
 
         try {
-            ProviderFuture request = new ProviderFuture(synchronization);
+            ProviderFuture request = provider.newProviderFuture(synchronization);
             requests.put(request, request);
             try {
                 provider.rollback(transactionInfo, nextTransactionId, request);
@@ -796,7 +882,7 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         checkClosedOrFailed();
 
         try {
-            ProviderFuture request = new ProviderFuture(synchronization);
+            ProviderFuture request = provider.newProviderFuture(synchronization);
             requests.put(request, request);
             try {
                 provider.recover(sessionId, request);
@@ -817,7 +903,7 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         checkClosedOrFailed();
 
         try {
-            ProviderFuture request = new ProviderFuture(synchronization);
+            ProviderFuture request = provider.newProviderFuture(synchronization);
             requests.put(request, request);
             try {
                 provider.pull(consumerId, timeout, request);
@@ -830,6 +916,14 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         }
     }
 
+    ProviderFuture newProviderFuture() {
+        return newProviderFuture(null);
+    }
+
+    ProviderFuture newProviderFuture(ProviderSynchronization synchronization) {
+        return provider.newProviderFuture(synchronization);
+    }
+
     //----- Property setters and getters -------------------------------------//
 
     @Override
@@ -907,6 +1001,14 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         connectionInfo.setValidatePropertyNames(validatePropertyNames);
     }
 
+    public boolean isValidateSelector() {
+        return connectionInfo.isValidateSelector();
+    }
+
+    public void setValidateSelector(boolean validateSelector) {
+        connectionInfo.setValidateSelector(validateSelector);
+    }
+
     public JmsPrefetchPolicy getPrefetchPolicy() {
         return connectionInfo.getPrefetchPolicy();
     }
@@ -1082,35 +1184,53 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         return connectionInfo.isUseDaemonThread();
     }
 
+    public boolean isCloseLinksThatFailOnReconnect() {
+        return connectionInfo.isCloseLinksThatFailOnReconnect();
+    }
+
+    public void setCloseLinksThatFailOnReconnect(boolean closeLinksThatFailOnReconnect) {
+        connectionInfo.setCloseLinksThatFailOnReconnect(closeLinksThatFailOnReconnect);
+    }
+
+    JmsTracer getTracer() {
+        return connectionInfo.getTracer();
+    }
+
     //----- Async event handlers ---------------------------------------------//
 
     @Override
     public void onInboundMessage(final JmsInboundMessageDispatch envelope) {
-
         JmsMessage incoming = envelope.getMessage();
         if (incoming != null) {
-            // Ensure incoming Messages are in readonly mode.
+            // Ensure incoming Messages are in read-only mode and configured properly
             incoming.setReadOnlyBody(true);
             incoming.setReadOnlyProperties(true);
-
             incoming.setValidatePropertyNames(isValidatePropertyNames());
         }
 
-        JmsMessageDispatcher dispatcher = sessions.get(envelope.getConsumerId().getParentId());
+        JmsMessageDispatcher dispatcher = null;
+
+        if (envelope.getConsumerInfo() != null && envelope.getConsumerInfo().getDispatcher() != null) {
+            dispatcher = envelope.getConsumerInfo().getDispatcher();
+        } else {
+            dispatcher = sessions.get(envelope.getConsumerId().getParentId());
+            if (dispatcher == null) {
+                dispatcher = connectionConsumers.get(envelope.getConsumerId());
+            }
+        }
+
         if (dispatcher != null) {
             dispatcher.onInboundMessage(envelope);
+        } else {
+            LOG.debug("Message inbound with no dispatcher registered for its consumer: {}", envelope.getConsumerId());
         }
 
         // Run the application callbacks on the connection executor to allow the provider to
         // return to its normal processing without waiting for client level processing to finish.
         if (!connectionListeners.isEmpty()) {
             for (final JmsConnectionListener listener : connectionListeners) {
-                executor.submit(new Runnable() {
-
-                    @Override
-                    public void run() {
-                        listener.onInboundMessage(envelope);
-                    }
+                executor.submit(() -> {
+                    listener.onInboundMessage(envelope);
                 });
             }
         }
@@ -1127,7 +1247,7 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
     }
 
     @Override
-    public void onFailedMessageSend(JmsOutboundMessageDispatch envelope, Throwable cause) {
+    public void onFailedMessageSend(JmsOutboundMessageDispatch envelope, ProviderException cause) {
         JmsSession session = sessions.get(envelope.getProducerId().getParentId());
         if (session != null) {
             session.onFailedMessageSend(envelope, cause);
@@ -1138,6 +1258,12 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
 
     @Override
     public void onConnectionInterrupted(final URI remoteURI) {
+        try {
+            LOG.info("Connection {} interrupted to server: {}", connectionInfo.getId(), URISupport.removeQuery(remoteURI));
+        } catch (URISyntaxException e) {
+            LOG.info("Connection {} interrupted to server: {}:{}", connectionInfo.getId(), remoteURI.getHost(), remoteURI.getPort());
+        }
+
         for (JmsSession session : sessions.values()) {
             session.onConnectionInterrupted();
         }
@@ -1159,16 +1285,25 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
     public void onConnectionRecovery(Provider provider) throws Exception {
         LOG.debug("Connection {} is starting recovery.", connectionInfo.getId());
 
-        ProviderFuture request = new ProviderFuture();
+        ProviderFuture request = provider.newProviderFuture();
         provider.create(connectionInfo, request);
         request.sync();
 
         for (JmsTemporaryDestination tempDestination : tempDestinations.values()) {
-            request = new ProviderFuture();
+            request = provider.newProviderFuture();
             provider.create(tempDestination, request);
             request.sync();
         }
 
+        for (JmsConnectionConsumer connectionConsumer : connectionConsumers.values()) {
+            JmsConsumerInfo consumerInfo = connectionConsumer.getConsumerInfo();
+            if (!consumerInfo.isClosed()) {
+                request = provider.newProviderFuture();
+                provider.create(consumerInfo, request);
+                request.sync();
+            }
+        }
+
         for (JmsSession session : sessions.values()) {
             session.onConnectionRecovery(provider);
         }
@@ -1181,6 +1316,15 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         setMessageFactory(provider.getMessageFactory());
         connectionInfo.setConnectedURI(provider.getRemoteURI());
 
+        for (JmsConnectionConsumer connectionConsumer : connectionConsumers.values()) {
+            JmsConsumerInfo consumerInfo = connectionConsumer.getConsumerInfo();
+            if (!consumerInfo.isClosed()) {
+                ProviderFuture request = provider.newProviderFuture();
+                provider.start(consumerInfo, request);
+                request.sync();
+            }
+        }
+
         for (JmsSession session : sessions.values()) {
             session.onConnectionRecovered(provider);
         }
@@ -1192,6 +1336,12 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
             session.onConnectionRestored();
         }
 
+        try {
+            LOG.info("Connection {} restored to server: {}", connectionInfo.getId(), URISupport.removeQuery(remoteURI));
+        } catch (URISyntaxException e) {
+            LOG.info("Connection {} restored to server: {}:{}", connectionInfo.getId(), remoteURI.getHost(), remoteURI.getPort());
+        }
+
         // Run the application callbacks on the connection executor to allow the provider to
         // return to its normal processing without waiting for client level processing to finish.
         for (final JmsConnectionListener listener : connectionListeners) {
@@ -1207,7 +1357,12 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
 
     @Override
     public void onConnectionEstablished(final URI remoteURI) {
-        LOG.info("Connection {} connected to remote Broker: {}", connectionInfo.getId(), remoteURI);
+        try {
+            LOG.info("Connection {} connected to server: {}", connectionInfo.getId(), URISupport.removeQuery(remoteURI));
+        } catch (URISyntaxException e) {
+            LOG.info("Connection {} connected to server: {}:{}", connectionInfo.getId(), remoteURI.getHost(), remoteURI.getPort());
+        }
+
         setMessageFactory(provider.getMessageFactory());
         connectionInfo.setConnectedURI(provider.getRemoteURI());
 
@@ -1225,13 +1380,21 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
     }
 
     @Override
-    public void onConnectionFailure(final IOException ex) {
+    public void onConnectionFailure(final ProviderException ex) {
         providerFailed(ex);
 
+        if (!closing.get() && !closed.get()) {
+            LOG.warn("Connection {} has failed due to: {}", connectionInfo.getId(), ex != null ? ex.getMessage() : "No error details provided.");
+        }
+
         // Signal that connection dropped we need to mark transactions as
         // failed, deliver failure events to asynchronous send completions etc.
         for (JmsSession session : sessions.values()) {
-            session.onConnectionInterrupted();
+            try {
+                session.onConnectionInterrupted();
+            } catch (Throwable t) {
+                LOG.warn("Exception while marking session interrupted", t);
+            }
         }
 
         onProviderException(ex);
@@ -1284,7 +1447,7 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
     }
 
     @Override
-    public void onResourceClosed(final JmsResource resource, final Throwable cause) {
+    public void onResourceClosed(final JmsResource resource, final ProviderException cause) {
         // Closure of the Connection itself is notified via onConnectionFailure
 
         // Run on the connection executor to free the provider to go do more work and avoid
@@ -1310,14 +1473,22 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
                     }
                 }
             } else if (resource instanceof JmsConsumerInfo) {
-                JmsSessionId parentId = ((JmsConsumerInfo) resource).getParentId();
-                JmsSession session = sessions.get(parentId);
-                if (session != null) {
-                    JmsMessageConsumer consumer = session.lookup((JmsConsumerId) resource.getId());
+                JmsConsumerInfo consumerInfo = (JmsConsumerInfo) resource;
+                if (consumerInfo.isConnectionConsumer()) {
+                    JmsConnectionConsumer consumer = connectionConsumers.get(consumerInfo.getId());
                     if (consumer != null) {
                         consumer.setFailureCause(cause);
                     }
-                }
+                } else {
+                    JmsSessionId parentId = consumerInfo.getParentId();
+                    JmsSession session = sessions.get(parentId);
+                    if (session != null) {
+                        JmsMessageConsumer consumer = session.lookup((JmsConsumerId) resource.getId());
+                        if (consumer != null) {
+                            consumer.setFailureCause(cause);
+                        }
+                    }
+                 }
             }
 
             executor.execute(new Runnable() {
@@ -1344,16 +1515,32 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
                             }
                         }
                     } else if (resource instanceof JmsConsumerInfo) {
-                        JmsSessionId parentId = ((JmsConsumerInfo) resource).getParentId();
-                        JmsSession session = sessions.get(parentId);
-                        if (session != null) {
-                            JmsMessageConsumer consumer = session.consumerClosed((JmsConsumerInfo) resource, cause);
-                            if (consumer != null) {
-                                for (JmsConnectionListener listener : connectionListeners) {
-                                    listener.onConsumerClosed(consumer, cause);
+                        JmsConsumerInfo consumerInfo = (JmsConsumerInfo) resource;
+                        if (consumerInfo.isConnectionConsumer()) {
+                            JmsConnectionConsumer consumer = connectionConsumers.get(consumerInfo.getId());
+                             if (consumer != null) {
+                                try {
+                                    if (consumer != null) {
+                                        consumer.shutdown(cause);
+                                    }
+                                } catch (Throwable error) {
+                                    LOG.trace("Ignoring exception thrown during cleanup of closed connection consumer", error);
                                 }
+
+                                onAsyncException(new JMSException("Connection Consumer remotely closed").initCause(cause));
                             }
-                        }
+                        } else {
+                            JmsSessionId parentId = consumerInfo.getParentId();
+                            JmsSession session = sessions.get(parentId);
+                            if (session != null) {
+                                JmsMessageConsumer consumer = session.consumerClosed((JmsConsumerInfo) resource, cause);
+                                if (consumer != null) {
+                                    for (JmsConnectionListener listener : connectionListeners) {
+                                        listener.onConsumerClosed(consumer, cause);
+                                    }
+                                 }
+                             }
+                         }
                     } else {
                         LOG.info("A JMS resource has been remotely closed: {}", resource);
                     }
@@ -1363,7 +1550,7 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
     }
 
     @Override
-    public void onProviderException(final Exception cause) {
+    public void onProviderException(final ProviderException cause) {
         // Report this to any registered exception listener, let the receiver
         // decide if it should be fatal.
         onAsyncException(cause);
@@ -1396,7 +1583,7 @@ public class JmsConnection implements AutoCloseable, Connection, TopicConnection
         }
     }
 
-    protected void providerFailed(IOException cause) {
+    protected void providerFailed(ProviderException cause) {
         failureCause.compareAndSet(null, cause);
     }
 }
